import abc
from typing import *

class ScpiTransportBase(object):
    __metaclass__ = abc.ABCMeta

    @classmethod
    @abc.abstractmethod
    def find_devices(cls, **kwargs):
        raise NotImplementedError("Please Implement this method")

    @abc.abstractmethod
    def __init__(self, resource, *args, **kwargs):
        super().__init__(resource, *args, **kwargs)

    @abc.abstractmethod
    def is_open(self) -> bool:
        raise NotImplementedError("Please Implement this method")

    def close(self) -> None:
        raise NotImplementedError("Please Implement this method")

    @abc.abstractmethod
    def reset(self) -> None:
        raise NotImplementedError("Please Implement this method")

    @abc.abstractmethod
    def readline(self, size: int = -1) -> str:
        raise NotImplementedError("Please Implement this method")

    @abc.abstractmethod
    def writeline(self, line: str) -> Optional[int]:
        raise NotImplementedError("Please Implement this method")


# TypeVar for type hinting subclasses of ScpiTransportBase
ScpiTransports = TypeVar('ScpiTransports', bound=ScpiTransportBase)


import logging
from typing import *
from .scpi_transport_base import ScpiTransportBase
from .scpi_transport_serial import ScpiSerialTransport
from .scpi_type_base import ScpiTypes
from .scpi_types import ScpiEvent
from .scpi_exceptions import ScpiErrorException

logger = logging.getLogger(__name__)


class ScpiDevice(object):
    def __init__(self, resource: str, transport_class: Type[ScpiTransportBase] = ScpiSerialTransport) -> None:
        self.resource = resource
        self.transport = transport_class(resource)
        logger.info(f'open [{transport_class.__name__}]: {resource}')

    def is_open(self):
        return self.transport.is_open

    def close(self) -> None:
        self.transport.close()

    def reset(self) -> None:
        self.transport.reset()

    def execute(self, header: str, param: Optional[ScpiTypes] = None,
                result: Optional[Type[ScpiTypes]] = type(None)) -> ScpiTypes:
        command = header
        resp = None
        response = None

        if param is not None:
            command = ' '.join([header, param.compose()])

        self.transport.writeline(command)

        if result is not type(None):
            response = result.parse(self.transport)
            resp = str(response)

        logger.info(f'exec [{type(param).__name__}] -> [{result.__name__ }]: <{command}> -> ({resp})')
        return response

    def check_error(self) -> Optional[ScpiEvent]:
        """Checks if an error is available in the device's SCPI error queue and returns it. Returns None otherwise"""
        scpi_error = self.execute('SYST:ERR:NEXT?', result=ScpiEvent)
        if scpi_error.code != 0:
            return scpi_error
        return None

    def raise_error(self) -> None:
        """Raise a ScpiErrorException when an scpi error is in the device's SCPI error queue"""
        scpi_error = self.check_error()
        if scpi_error is not None:
            raise ScpiErrorException(scpi_error)



from .scpi_types import ScpiEvent

class ScpiException(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)

class ScpiErrorException(ScpiException):
    def __init__(self, scpi_error: ScpiEvent = None):
        self.scpi_error = scpi_error
        super().__init__(self, str(self))

    def __str__(self):
        return f'SCPI Error {self.scpi_error.code}: {self.scpi_error.description}' \
               + f' ({self.scpi_error.info})' if self.scpi_error.info is not None else ''


import serial
import logging
from typing import *
from .scpi_transport_base import ScpiTransportBase

logger = logging.getLogger(__name__)

class ScpiSerialTransport(serial.Serial, ScpiTransportBase):
    @classmethod
    def find_devices(cls, regexp: str = ''):
        import serial.tools.list_ports

        resources = [a[0] for a in serial.tools.list_ports.grep(regexp)]
        return resources

    def __init__(self, port, timeout = 5):
        super().__init__(port = port, timeout = timeout, exclusive=True)
        self.reset()

    def reset(self) -> None:
        self.reset_input_buffer()
        self.reset_output_buffer()
        self.timeout = 0
        while(self.read() != b''):
            pass
        self.timeout = 5

    def readline(self, size: int = -1) -> str:
        size = None if size == -1 else size
        data = super().read_until(b'\r\n', size=size)
        line = data.decode('ascii').strip('\r\n')
        logger.debug(f'read: {line}<CRLF>')
        return line

    def writeline(self, line: str) -> Optional[int]:
        logger.debug(f'write: {line}<CRLF>')
        data = (line + '\r\n').encode('ascii')
        return super().write(data)